@@ -1,12 +1,19 @@
 import dataclasses
-import json
+import os
 import time
 from enum import Enum
 from typing import List, Optional
-from urllib.parse import urlencode, urljoin
+from urllib.parse import urljoin
 from uuid import UUID
 
 import requests
+from urllib3.exceptions import InsecureRequestWarning
+
+verify_ssl = os.environ.get("REQUESTS_VERIFY_SSL", "true")
+
+verify_ssl = verify_ssl.lower() not in ["false", "no", "0"]
+
+requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
 
 
 def is_uuid(value):
@@ -65,6 +72,37 @@ class ResourceReportRecord:
     body: str
 
 
+class OfferingComponentBillingTypes:
+    FIXED = "fixed"
+    USAGE = "usage"
+    ONE_TIME = "one"
+    ON_PLAN_SWITCH = "few"
+    LIMIT = "limit"
+
+
+class OfferingComponentLimitPeriod:
+    MONTH = "month"
+    ANNUAL = "annual"
+    TOTAL = "total"
+
+
+@dataclasses.dataclass
+class OfferingComponent:
+    # The required fields are billing_type, type, name and measured_unit only
+    billing_type: str
+    type: str
+    name: str
+    measured_unit: str
+    description: str = ""
+    limit_period: str = ""
+    limit_amount: int = None
+    article_code: str = ""
+    max_value: int = None
+    min_value: int = None
+    is_boolean: bool = None
+    default_limit: int = None
+
+
 class ResourceState(Enum):
     OK = "ok"
     ERRED = "erred"
@@ -77,6 +115,16 @@ class ProjectRole(Enum):
     MEMBER = "member"
 
 
+class SlurmAllocationState(Enum):
+    CREATING = "creating"
+    UPDATE_SCHEDULED = "updating"
+    UPDATING = "deletion_scheduled"
+    DELETION_SCHEDULED = "update_scheduled"
+    DELETING = "deleting"
+    OK = "ok"
+    ERRED = "erred"
+
+
 class PaymentProfileType(Enum):
     FIXED_PRICE = "fixed_price"
     MONTHLY_INVOICES = "invoices"
@@ -97,9 +145,11 @@ class WaldurClient(object):
         Customers = "customers"
         Flavor = "openstacktenant-flavors"
         FloatingIP = "openstacktenant-floating-ips"
+        FreeIPAProfiles = "freeipa-profiles"
         Image = "openstacktenant-images"
         Instance = "openstacktenant-instances"
         Invoice = "invoices"
+        InvoiceItems = "invoice-items"
         MarketplaceCategories = "marketplace-categories"
         MarketplaceOffering = "marketplace-offerings"
         MarketplaceOrder = "marketplace-orders"
@@ -107,6 +157,8 @@ class WaldurClient(object):
         MarketplacePlan = "marketplace-plans"
         MarketplaceResources = "marketplace-resources"
         MarketplaceStats = "marketplace-stats"
+        MarketplaceSlurm = "marketplace-slurm"
+        MarketplaceSlurmRemote = "marketplace-slurm-remote"
         OfferingPermissions = "marketplace-offering-permissions"
         OfferingUsers = "marketplace-offering-users"
         PaymentProfiles = "payment-profiles"
@@ -117,6 +169,9 @@ class WaldurClient(object):
         RemoteEduteams = "remote-eduteams"
         SecurityGroup = "openstacktenant-security-groups"
         ServiceProviders = "marketplace-service-providers"
+        SlurmAllocations = "slurm-allocations"
+        SlurmAssociations = "slurm-associations"
+        SlurmAllocationUserUsages = "slurm-allocation-user-usages"
         Snapshot = "openstacktenant-snapshots"
         SshKey = "keys"
         Subnet = "openstacktenant-subnets"
@@ -126,6 +181,9 @@ class WaldurClient(object):
         Users = "users"
         Volume = "openstacktenant-volumes"
         VolumeType = "openstacktenant-volume-types"
+        ServerGroup = "openstacktenant-server-groups"
+        SupportIssues = "support-issues"
+        SupportComments = "support-comments"
 
     marketplaceScopeEndpoints = {
         "OpenStackTenant.Instance": Endpoints.Instance,
@@ -161,11 +219,18 @@ class WaldurClient(object):
         return self._build_url("/".join(parts))
 
     def _parse_error(self, response):
+        reason = getattr(response, "reason")
+
         try:
-            reason = response.json()
+            reason = "%s. %s" % (reason, response.json())
         except ValueError:
-            reason = "Unable to parse JSON"
-        details = "Status: %s. Reason: %s." % (response.status_code, reason)
+            pass
+
+        details = "URL %s. Status: %s. Reason: %s." % (
+            response.url,
+            response.status_code,
+            reason,
+        )
         return "Server refuses to communicate. %s" % details
 
     def _make_request(self, method, url, valid_states, retry_count=3, **kwargs):
@@ -174,7 +239,7 @@ class WaldurClient(object):
                 "Reached a limit of retries for the operation: %s %s" % (method, url)
             )
 
-        params = dict(headers=self.headers)
+        params = dict(headers=self.headers, verify=verify_ssl)
         params.update(kwargs)
 
         try:
@@ -242,13 +307,16 @@ class WaldurClient(object):
         return self._make_request("head", url, valid_states=[200], **kwargs)
 
     def _post(self, url, valid_states, **kwargs):
-        return self._make_request("post", url, valid_states, 3, **kwargs)
+        return self._make_request("post", url, valid_states, **kwargs)
 
     def _put(self, url, valid_states, **kwargs):
-        return self._make_request("put", url, valid_states, 3, **kwargs)
+        return self._make_request("put", url, valid_states, **kwargs)
+
+    def _patch(self, url, valid_states, **kwargs):
+        return self._make_request("patch", url, valid_states, **kwargs)
 
     def _delete(self, url, valid_states, **kwargs):
-        return self._make_request("delete", url, valid_states, 3, **kwargs)
+        return self._make_request("delete", url, valid_states, **kwargs)
 
     def _make_get_query(self, url, query_params, get_first=False, get_few=False):
         """
@@ -332,7 +400,11 @@ class WaldurClient(object):
 
     def _update_resource(self, endpoint, uuid, payload):
         url = self._build_resource_url(endpoint, uuid)
-        return self._put(url, [200], data=json.dumps(payload))
+        return self._put(url, [200], json=payload)
+
+    def _patch_resource(self, endpoint, uuid, payload):
+        url = self._build_resource_url(endpoint, uuid)
+        return self._patch(url, [200], json=payload)
 
     def _delete_resource_by_url(self, url):
         return self._delete(url, [202, 204])
@@ -354,9 +426,11 @@ class WaldurClient(object):
     def get_user(self, identifier):
         return self._get_resource(self.Endpoints.Users, identifier)
 
-    def list_users(self):
-        url = self._build_url(self.Endpoints.Users)
-        return self._get_all(url)
+    def list_users(self, filters=None):
+        return self._query_resource_list(self.Endpoints.Users, filters)
+
+    def list_freeipa_profiles(self, filters=None):
+        return self._query_resource_list(self.Endpoints.FreeIPAProfiles, filters)
 
     def count_users(self, **kwargs):
         url = self._build_url(self.Endpoints.Users)
@@ -398,6 +472,9 @@ class WaldurClient(object):
             self.Endpoints.SecurityGroup, identifier, settings_uuid
         )
 
+    def _get_server_group(self, identifier, settings_uuid):
+        return self._get_property(self.Endpoints.ServerGroup, identifier, settings_uuid)
+
     def _get_floating_ip(self, address):
         return self._query_resource(self.Endpoints.FloatingIP, {"address": address})
 
@@ -623,7 +700,10 @@ class WaldurClient(object):
                 raise ValidationError(
                     "You should specify project name if name is not UUID"
                 )
-            query = {"project_name": project, "name_exact": name}
+            if is_uuid(project):
+                query = {"project_uuid": project, "name_exact": name}
+            else:
+                query = {"project_name": project, "name_exact": name}
             return self._query_resource(endpoint, query)
 
     def get_instance(self, name, project=None):
@@ -684,6 +764,9 @@ class WaldurClient(object):
             WaldurClient.Endpoints.MarketplaceResources, resource_uuid
         )
 
+    def filter_marketplace_resources(self, filters=None):
+        return self._query_resource_list(self.Endpoints.MarketplaceResources, filters)
+
     def list_marketplace_resources(
         self,
         provider_uuid: str = None,
@@ -777,14 +860,6 @@ class WaldurClient(object):
         if wait:
             self._wait_for_resource(self.Endpoints.Instance, uuid, interval, timeout)
 
-    def delete_instance(self, uuid, delete_volumes=True, release_floating_ips=True):
-        base_url = self._build_resource_url(self.Endpoints.Instance, uuid)
-        params = dict(
-            delete_volumes=delete_volumes, release_floating_ips=release_floating_ips
-        )
-        url = base_url + "?" + urlencode(params)
-        return self._delete_resource_by_url(url)
-
     def update_instance_security_groups(
         self,
         instance_uuid,
@@ -834,9 +909,6 @@ class WaldurClient(object):
         uuid = volume["uuid"]
         return self._update_resource(self.Endpoints.Volume, uuid, payload)
 
-    def delete_volume(self, uuid):
-        return self._delete_resource(self.Endpoints.Volume, uuid)
-
     def detach_volume(self, uuid, wait=True, interval=10, timeout=600):
         """
         Detach OpenStack volume from instance and wait until operation is completed.
@@ -1030,6 +1102,49 @@ class WaldurClient(object):
         )
         return self._post(url, valid_states=[200])
 
+    def marketplace_order_item_terminate(self, order_item_uuid: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOrderItem,
+            order_item_uuid,
+            action="terminate",
+        )
+
+        return self._post(url, valid_states=[202])
+
+    def marketplace_order_item_set_state_executing(self, order_item_uuid: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOrderItem,
+            order_item_uuid,
+            action="set_state_executing",
+        )
+
+        return self._post(url, valid_states=[200])
+
+    def marketplace_order_item_set_state_done(self, order_item_uuid: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOrderItem,
+            order_item_uuid,
+            action="set_state_done",
+        )
+
+        return self._post(url, valid_states=[200])
+
+    def marketplace_order_item_set_state_erred(
+        self,
+        order_item_uuid: str,
+        error_message: str = "",
+        error_traceback: str = "",
+    ):
+        payload = {"error_message": error_message, "error_traceback": error_traceback}
+
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOrderItem,
+            order_item_uuid,
+            action="set_state_erred",
+        )
+
+        return self._post(url, json=payload, valid_states=[200])
+
     def _get_resource_from_creation_order(
         self,
         order_uuid,
@@ -1135,6 +1250,7 @@ class WaldurClient(object):
         ssh_key=None,
         data_volume_size=None,
         security_groups=None,
+        server_group=None,
         tags=None,
         user_data=None,
         check_mode=False,
@@ -1163,6 +1279,7 @@ class WaldurClient(object):
             No data volume is going to be created if empty.
         :param data_volume_type: UUID or name of data volume type.
         :param security_groups: list of security groups to add to the instance.
+        :param server_group: A server group to add to the instance.
         :param tags: list of tags to add to the instance.
         :param user_data: additional data that will be added to the instance.
         :return: an instance as a dictionary.
@@ -1211,6 +1328,9 @@ class WaldurClient(object):
         if data_volume_type:
             volume_type = self._get_volume_type(data_volume_type, settings_uuid)
             attributes.update({"data_volume_type": volume_type["url"]})
+        if server_group:
+            server_group = self._get_resource(self.Endpoints.ServerGroup, server_group)
+            attributes.update({"server_group": server_group["url"]})
 
         resource_uuid = self._create_scope_via_marketplace(
             offering["uuid"],
@@ -1405,22 +1525,14 @@ class WaldurClient(object):
 
     def _serialize_project(
         self,
-        name=None,
-        backend_id=None,
-        description="",
-        end_date=None,
-        oecd_fos_2007_code=None,
         type_uuid=None,
+        **kwargs,
     ):
         type_url = type_uuid and self._build_resource_url(
             self.Endpoints.ProjectTypes, type_uuid
         )
         return {
-            "name": name,
-            "backend_id": backend_id,
-            "description": description,
-            "end_date": end_date,
-            "oecd_fos_2007_code": oecd_fos_2007_code,
+            **kwargs,
             "type": type_url,
         }
 
@@ -1434,7 +1546,7 @@ class WaldurClient(object):
 
     def update_project(self, project_uuid, **kwargs):
         payload = self._serialize_project(**kwargs)
-        return self._update_resource(
+        return self._patch_resource(
             self.Endpoints.Project, project_uuid, payload=payload
         )
 
@@ -1458,7 +1570,13 @@ class WaldurClient(object):
         )
 
     def marketplace_resource_create_order(
-        self, project_uuid, offering_uuid, plan_uuid=None, attributes=None, limits=None
+        self,
+        project_uuid,
+        offering_uuid,
+        plan_uuid=None,
+        attributes=None,
+        limits=None,
+        callback_url=None,
     ):
         attributes = attributes or {}
         limits = limits or {}
@@ -1475,6 +1593,9 @@ class WaldurClient(object):
                 self.Endpoints.MarketplacePlan, plan_uuid
             )
 
+        if callback_url:
+            order_item["callback_url"] = callback_url
+
         # TODO: replace with checkbox data from frontend
         order_item["accepting_terms_of_service"] = True
 
@@ -1484,16 +1605,26 @@ class WaldurClient(object):
         }
         return self._create_resource(self.Endpoints.MarketplaceOrder, payload=payload)
 
-    def marketplace_resource_update_limits_order(self, resource_uuid, limits):
+    def marketplace_resource_update_limits_order(
+        self, resource_uuid, limits, callback_url=None
+    ):
         payload = {"limits": limits}
+        if callback_url:
+            payload["callback_url"] = callback_url
         url = self._build_resource_url(
             self.Endpoints.MarketplaceResources, resource_uuid, action="update_limits"
         )
         return self._post(url, valid_states=[200], json=payload)["order_uuid"]
 
-    def marketplace_resource_terminate_order(self, resource_uuid, options=None):
+    def marketplace_resource_terminate_order(
+        self, resource_uuid, options=None, callback_url=None
+    ):
         if options:
             options = {"attributes": options}
+        if callback_url:
+            if not options:
+                options = {}
+            options["callback_url"] = callback_url
         url = self._build_resource_url(
             self.Endpoints.MarketplaceResources, resource_uuid, action="terminate"
         )
@@ -1545,6 +1676,9 @@ class WaldurClient(object):
         url = self._build_resource_url(self.Endpoints.Invoice, invoice_uuid, "paid")
         return self._post(url, valid_states=[200])
 
+    def list_invoice_items(self, filters=None):
+        return self._query_resource_list(self.Endpoints.InvoiceItems, filters)
+
     def list_payment_profiles(self, filters=None):
         if "payment_type" in filters:
             filters["payment_type"] = filters["payment_type"].value
@@ -1736,12 +1870,11 @@ class WaldurClient(object):
         }
         return self._post(endpoint, valid_states=[201], json=payload)
 
-    def list_remote_offering_users(self, filters):
+    def list_remote_offering_users(self, filters=None):
         return self._query_resource_list(self.Endpoints.OfferingUsers, filters)
 
-    def list_service_providers(self, filters):
-        endpoint = self._build_url(self.Endpoints.ServiceProviders)
-        return self._query_resource_list(endpoint, filters)
+    def list_service_providers(self, filters=None):
+        return self._query_resource_list(self.Endpoints.ServiceProviders, filters)
 
     def list_service_provider_users(self, service_provider_uuid):
         endpoint = self._build_resource_url(
@@ -1773,6 +1906,123 @@ class WaldurClient(object):
         endpoint = self._build_url(self.Endpoints.MarketplaceStats, endpoint)
         return self._make_request("get", endpoint, valid_states=[200])
 
+    def get_slurm_allocation(self, uuid: str):
+        if not is_uuid(uuid):
+            raise ValidationError(
+                "The UUID of SLURM allocation has unexpected format: %s" % uuid
+            )
+        return self._get_resource(self.Endpoints.SlurmAllocations, uuid)
+
+    def list_slurm_allocations(self, filters=None):
+        return self._query_resource_list(self.Endpoints.SlurmAllocations, filters)
+
+    def set_slurm_allocation_state(
+        self, marketplace_resource_uuid: str, state: SlurmAllocationState
+    ):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceSlurmRemote,
+            marketplace_resource_uuid,
+            "set_state",
+        )
+        payload = {"state": state.value}
+        self._post(url, valid_states=[200], json=payload)
+
+    def set_slurm_allocation_backend_id(
+        self, marketplace_resource_uuid: str, backend_id: str
+    ):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceSlurmRemote,
+            marketplace_resource_uuid,
+            "set_backend_id",
+        )
+        payload = {"backend_id": backend_id}
+        self._post(url, valid_states=[200], json=payload)
+
+    def list_slurm_associations(self, filters=None):
+        return self._query_resource_list(self.Endpoints.SlurmAssociations, filters)
+
+    def create_slurm_association(self, marketplace_resource_uuid: str, username: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceSlurmRemote,
+            marketplace_resource_uuid,
+            "create_association",
+        )
+        payload = {"username": username}
+        return self._post(url, valid_states=[200, 201], json=payload)
+
+    def delete_slurm_association(self, marketplace_resource_uuid: str, username: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceSlurmRemote,
+            marketplace_resource_uuid,
+            "delete_association",
+        )
+        payload = {"username": username}
+        return self._post(url, valid_states=[200], json=payload)
+
+    def set_slurm_allocation_limits(
+        self,
+        marketplace_resource_uuid: str,
+        limits: dict,
+    ):
+        if not is_uuid(marketplace_resource_uuid):
+            raise ValidationError(
+                "The UUID of marketplace resource has unexpected format: %s"
+                % marketplace_resource_uuid
+            )
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceSlurmRemote,
+            marketplace_resource_uuid,
+            "set_limits",
+        )
+        return self._post(url, valid_states=[200], json=limits)
+
+    def list_slurm_allocation_user_usage(self, filters=None):
+        return self._query_resource_list(
+            self.Endpoints.SlurmAllocationUserUsages, filters
+        )
+
+    def update_offering_components(
+        self, offering_uuid: str, components: List[OfferingComponent]
+    ):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOffering, offering_uuid, "update_components"
+        )
+        components_json = [
+            {
+                k: v for k, v in dataclasses.asdict(component).items() if v
+            }  # Drop all keys with None or empty value
+            for component in components
+        ]
+        return self._post(url, valid_states=[200], json=components_json)
+
+    def create_support_issue(
+        self, summary, issue_type, caller_url, remote_id, **kwargs
+    ):
+        payload = {
+            "summary": summary,
+            "type": issue_type,
+            "caller": caller_url,
+            "remote_id": remote_id,
+        }
+        payload.update(kwargs)
+        return self._create_resource(self.Endpoints.SupportIssues, payload=payload)
+
+    def list_support_issues(self, filters=None):
+        return self._query_resource_list(self.Endpoints.SupportIssues, filters)
+
+    def list_support_comments(self, filters=None):
+        return self._query_resource_list(self.Endpoints.SupportComments, filters)
+
+    def create_support_comments(self, issue_uuid, description, remote_id, **kwargs):
+        payload = {
+            "description": description,
+            "remote_id": remote_id,
+        }
+        payload.update(kwargs)
+        return self._create_resource(
+            self.Endpoints.SupportIssues + f"/{issue_uuid}/comment/", payload=payload
+        )
+
 
 def waldur_full_argument_spec(**kwargs):
     spec = dict(
